import inspect
import os
import sys

from sphinx.ext import apidoc

from minimalkv import __version__ as version

release = version

sys.path.append("../")

package = "minimalkv"
html_theme = "alabaster"

__location__ = os.path.join(
    os.getcwd(),
    os.path.dirname(inspect.getfile(inspect.currentframe())),  # type: ignore
)


# Generate module references
output_dir = os.path.abspath(os.path.join(__location__, "../docs/_rst"))
module_dir = os.path.abspath(os.path.join(__location__, "..", package))

apidoc_parameters = ["-f", "-e", "-o", output_dir, module_dir]
apidoc.main(apidoc_parameters)

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.intersphinx",
]

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"

# General information about the project.
project = "minimalkv"
copyright = "2011-2021, The minimalkv contributors"


exclude_trees = ["_build"]

pygments_style = "sphinx"

# intersphinx_mapping = {
#    "http://docs.python.org/": None,
#    "http://docs.pythonboto.org/en/latest/": None,
#    "http://sendapatch.se/projects/pylibmc/": None,
#    "http://www.sqlalchemy.org/docs/": None,
#    "http://redis-py.readthedocs.org/en/latest/": None,
# }



import os
import time
from logging import getLogger
from random import randint
from typing import Union
from urllib.parse import quote_plus

import pytest
from boto3 import Session

from minimalkv import get_store_from_url

logger = getLogger(__name__)


@pytest.fixture()
def aws_credentials() -> tuple[str, str, Union[str, None]]:
    env_var_name = "AWS_PROFILE"
    profile_name = os.environ.get(env_var_name, None)

    access_key: Union[str, None] = None
    secret_key: Union[str, None] = None
    session_token: Union[str, None] = None

    if profile_name:
        session = Session(profile_name=profile_name)
        aws_credentials = session.get_credentials()
        assert aws_credentials is not None
        access_key = aws_credentials.access_key
        secret_key = aws_credentials.secret_key
        session_token = aws_credentials.token
    else:
        access_key = os.environ.get("AWS_ACCESS_KEY_ID", None)
        secret_key = os.environ.get("AWS_SECRET_ACCESS_KEY", None)
        session_token = os.environ.get("AWS_SESSION_TOKEN", None)

    if not (access_key and secret_key):
        msg = "No s3 credentials available. "

        if "CI_IN_FORK" not in os.environ or os.environ["CI_IN_FORK"].lower() == "true":
            # We skip if the variable is not set at all (local development)
            # or if it is explicitly set to "true".

            if "CI_IN_FORK" in os.environ:
                msg += "Skipping, because you (or the CI) set 'CI_IN_FORK=true'."
            else:
                # If running
                msg += (
                    "If you want to execute this integration test, "
                    f"set '{env_var_name}' env variable to "
                    "provide a valid AWS profile or set 'AWS_ACCESS_KEY_ID' and "
                    "'AWS_SECRET_ACCESS_KEY' and optional 'AWS_SESSION_TOKEN'."
                )

            pytest.skip(reason=msg)

        # If in CI of base repo and credentials couldn't be acquired, fail test

    assert access_key and secret_key
    return (
        access_key,
        secret_key,
        session_token,
    )


@pytest.fixture()
def ci_bucket_name() -> str:
    return "minimalkv-test-ci-bucket"


@pytest.fixture()
def ci_s3_point() -> str:
    return "s3.eu-north-1.amazonaws.com"


@pytest.fixture()
def role_arn() -> str:
    return "arn:aws:iam::211125346859:role/S3MinimalKvAccessRole"


def get_s3_url(
    access_key: str,
    secret_key: str,
    session_token: Union[str, None],
    bucket_name: str,
    s3_point: str,
) -> str:
    access_key = quote_plus(access_key)
    secret_key = quote_plus(secret_key)
    return f"hs3://{access_key}:{secret_key}@{s3_point}/{bucket_name}?force_bucket_suffix=false&create_if_missing=false&session_token={session_token}"


@pytest.fixture()
def test_id() -> str:
    return f"test-id-{randint(1, 1000)}"


def test_s3fs_aws_integration(
    test_id,
    aws_credentials: tuple[str, str, Union[str, None]],
    ci_bucket_name,
    ci_s3_point,
):
    """Authenticates with AWS S3 bucket via short-lived credentials and tests basic operation of S3FSStore.

    Test the basic interface:
    - keys()
    - put()
    - get()
    - delete()
    """
    access_key, secret_key, session_token = aws_credentials

    print(
        f"Testing with access_key: {access_key}, secret_key: {secret_key}, session_token: {session_token}"
    )

    bucket = get_store_from_url(
        get_s3_url(access_key, secret_key, session_token, ci_bucket_name, ci_s3_point)
    )

    new_filename = f"{test_id}-folder/file"  # Testing the h of hs3
    new_content = b"content"

    assert new_filename not in bucket.keys(), "Test prerequisites not fulfilled."

    bucket.put(new_filename, new_content)
    assert new_filename in bucket.keys()
    assert bucket.get(new_filename) == new_content

    bucket.delete(new_filename)
    assert new_filename not in bucket.keys()


@pytest.mark.slow
def test_sts_refresh_mechanism(
    aws_credentials: tuple[str, str, Union[str, None]],
    ci_bucket_name,
    ci_s3_point,
    role_arn,
):
    """
    In an enterprise environment, we usually have static sts credentials that can be used to assume a role with necessary permissions.

    To verify that a refreshing mechanism works, we use the following approach:
    - Use the `is_sts_credentials` parameter to signal that the credentials are STS credentials.
    - Under the hood, minimalkv will assume the specified role (via `sts_assume_role__RoleArn`)
    and use the credentials to access the bucket.
    - Test whether a basic operation works. Wait for 15 minutes until the credentials expire.
    - Test whether the basic operation still works.
    """
    access_key, secret_key, session_token = aws_credentials
    role_arn = "arn:aws:iam::211125346859:role/S3MinimalKvAccessRole"
    base_url = get_s3_url(
        access_key, secret_key, session_token, ci_bucket_name, ci_s3_point
    )
    sts_url_extension = f"is_sts_credentials=true&sts_assume_role__RoleArn={quote_plus(role_arn)}&sts_assume_role__DurationSeconds=900"

    bucket = get_store_from_url(f"{base_url}&{sts_url_extension}")

    bucket.iter_keys()

    for i in range(15):
        logger.info(
            f"Sleeping for 60 seconds ({i+1}/15) to wait for credentials to expire."
        )
        time.sleep(60)

    time.sleep(10)  # Some buffer time

    bucket.iter_keys()  # no assert as this will raise if the credentials are expired



from collections.abc import Iterator
from io import BytesIO
from typing import BinaryIO

from sqlalchemy import Column, LargeBinary, String, Table, exists, select
from sqlalchemy.orm import Session

from minimalkv import CopyMixin, KeyValueStore


class SQLAlchemyStore(KeyValueStore, CopyMixin):  # noqa D
    def __init__(self, bind, metadata, tablename):
        self.bind = bind

        self.table = Table(
            tablename,
            metadata,
            # 250 characters is the maximum key length that we guarantee can be
            # handled by any kind of backend
            Column("key", String(250), primary_key=True),
            Column("value", LargeBinary, nullable=False),
        )

    def _has_key(self, key: str) -> bool:
        with Session(self.bind) as session:
            return session.query(exists().where(self.table.c.key == key)).scalar()

    def _delete(self, key: str) -> None:
        with Session(self.bind) as session:
            session.execute(self.table.delete().where(self.table.c.key == key))
            session.commit()

    def _get(self, key: str) -> bytes:
        with Session(self.bind) as session:
            stmt = select(self.table.c.value).where(self.table.c.key == key)
            rv = session.execute(stmt).scalar()

            if not rv:
                raise KeyError(key)

            return rv

    def _open(self, key: str) -> BinaryIO:
        return BytesIO(self._get(key))

    def _copy(self, source: str, dest: str):
        with Session(self.bind) as session:
            # Find data at source key
            data_to_copy = self._get(source)

            # delete the potential existing previous key
            self.delete(dest)

            # insert new
            session.execute(
                self.table.insert().values(
                    {
                        "key": dest,
                        "value": data_to_copy,
                    }
                )
            )
            session.commit()
        return dest

    def _put(self, key: str, data: bytes) -> str:
        with Session(self.bind) as session:
            # delete the old
            session.execute(self.table.delete().where(self.table.c.key == key))

            # insert new
            session.execute(
                self.table.insert().values(
                    {
                        "key": key,
                        "value": data,
                    }
                )
            )
            session.commit()
        return key

    def _put_file(self, key: str, file: BinaryIO) -> str:
        return self._put(key, file.read())

    def iter_keys(self, prefix: str = "") -> Iterator[str]:  # noqa D
        with Session(self.bind) as session:
            query = select(self.table.c.key)
            if prefix != "":
                query = query.where(self.table.c.key.like(prefix + "%"))
            return (str(v[0]) for v in session.execute(query))



import os
from typing import NamedTuple

import pytest

from minimalkv import get_store_from_url


class User(NamedTuple):
    access_key: str
    secret_key: str
    bucket_name: str
    # The mapping between users and accessible buckets can be arbitrary.
    # The config in "tests/minio-container/policies" is very simple, allowing
    # user1 -> bucket1
    # user2 -> bucket2
    # Including one bucket_name per `User` here is just for readability

    def get_store_from_config(self):
        return get_store_from_url(
            f"hs3://{self.access_key}:{self.secret_key}@localhost:9000/{self.bucket_name}?force_bucket_suffix=false&verify=false"
        )


user1 = User("user1", "password1", "bucket1")
user2 = User("user2", "password2", "bucket2")


def test_access_multiple_users():
    """Verify that two buckets of different users can be accessed in the same python process."""
    bucket1 = user1.get_store_from_config()

    assert bucket1.keys() == ["file1.txt"]

    bucket2 = user2.get_store_from_config()

    assert bucket2.keys() == ["file2.txt"]


def test_example_interaction():
    """Tests basic operation of S3FSStore.

    Test basic interface:
    - keys()
    - put()
    - get()
    - delete()
    """
    bucket = user1.get_store_from_config()

    new_filename = "some-non-existing-file"
    new_content = b"content"

    assert new_filename not in bucket.keys(), "Test prerequisites not fulfilled."

    bucket.put(new_filename, new_content)
    assert new_filename in bucket.keys()
    assert bucket.get(new_filename) == new_content

    bucket.delete(new_filename)
    assert new_filename not in bucket.keys()


@pytest.fixture
def clean_env(monkeypatch):
    # Important because another test sets the environment variables when accessing
    # the bucket.
    monkeypatch.setattr(os, "environ", {})
    yield


def test_no_env_side_effects():
    pre_env_state = os.environ.copy()

    bucket = user1.get_store_from_config()

    assert dict(os.environ) == dict(
        pre_env_state
    ), "Retrieved bucket should not modify the environment."

    bucket.keys()

    assert (
        os.environ == pre_env_state
    ), "Performing operations on the bucket should not modify the environment."



import os
from contextlib import contextmanager

import pytest

boto = pytest.importorskip("boto", reason="'boto' is not available")
from io import BytesIO

from basic_store import BasicStore
from bucket_manager import boto_credentials, uuid
from conftest import ExtendedKeyspaceTests
from url_store import UrlStore

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.net.botostore import BotoStore


@contextmanager
def boto_bucket(
    access_key,
    secret_key,
    host,
    connect_func="connect_s3",
    ordinary_calling_format=False,
    bucket_name=None,
    port=None,
    is_secure=True,
):
    if ordinary_calling_format:
        from boto.s3.connection import OrdinaryCallingFormat

        conn = getattr(boto, connect_func)(
            access_key,
            secret_key,
            host=host,
            calling_format=OrdinaryCallingFormat(),
            port=port,
            is_secure=is_secure,
        )
    else:
        conn = getattr(boto, connect_func)(
            access_key, secret_key, host=host, port=port, is_secure=is_secure
        )

    name = bucket_name or f"testrun-bucket-{uuid()}"
    bucket = conn.create_bucket(name)

    yield bucket

    for key in bucket.list():
        key.delete()
    bucket.delete()


@pytest.fixture(
    params=boto_credentials, ids=[c["access_key"] for c in boto_credentials]
)
def credentials(request):
    return request.param


@pytest.fixture
def bucket(credentials):
    with boto_bucket(**credentials) as bucket:
        yield bucket


class TestBotoStorage(BasicStore, UrlStore):
    @pytest.fixture(params=[True, False])
    def reduced_redundancy(self, request):
        return request.param

    @pytest.fixture
    def storage_class(self, reduced_redundancy):
        return "REDUCED_REDUNDANCY" if reduced_redundancy else "STANDARD"

    @pytest.fixture(params=["", "/test-prefix"])
    def prefix(self, request):
        return request.param

    @pytest.fixture
    def store(self, bucket, prefix, reduced_redundancy):
        return BotoStore(bucket, prefix, reduced_redundancy=reduced_redundancy)

    # Disable max key length test as it leads to problems with minio
    test_max_key_length = None

    def test_get_filename_nonexistent(self, store, key, tmp_path):
        # NOTE: boto misbehaves here and tries to erase the target file
        # the parent tests use /dev/null, which you really should not try
        # to os.remove!
        with pytest.raises(KeyError):
            store.get_file(key, os.path.join(str(tmp_path), "a"))

    def test_key_error_on_nonexistent_get_filename(self, store, key, tmp_path):
        with pytest.raises(KeyError):
            store.get_file(key, os.path.join(str(tmp_path), "a"))

    def test_storage_class_put(self, store, prefix, key, value, storage_class, bucket):
        store.put(key, value)

        keyname = prefix + key

        if storage_class != "STANDARD":
            pytest.xfail("boto does not support checking the storage class?")

        assert bucket.lookup(keyname).storage_class == storage_class

    def test_storage_class_putfile(
        self, store, prefix, key, value, storage_class, bucket
    ):
        store.put_file(key, BytesIO(value))

        keyname = prefix + key

        if storage_class != "STANDARD":
            pytest.xfail("boto does not support checking the storage class?")
        assert bucket.lookup(keyname).storage_class == storage_class


class TestExtendedKeyspaceBotoStore(TestBotoStorage, ExtendedKeyspaceTests):
    @pytest.fixture
    def store(self, bucket, prefix, reduced_redundancy):
        class ExtendedKeyspaceStore(ExtendedKeyspaceMixin, BotoStore):
            pass

        return ExtendedKeyspaceStore(
            bucket, prefix, reduced_redundancy=reduced_redundancy
        )



import re
import time
from collections.abc import Iterator
from io import BytesIO
from typing import BinaryIO, Optional, Union

from dulwich.objects import Blob, Commit, Tree
from dulwich.repo import Repo

from minimalkv import __version__
from minimalkv._key_value_store import KeyValueStore


def _on_tree(
    repo: Repo,
    tree: Tree,
    components: list[bytes],
    obj: Optional[Union[Blob, Tree]],
) -> list[Tree]:
    """Mount an object on a tree, using the given path components.

    Parameters
    ----------
    repo : dulwich.objects.Repo
        Repository.
    tree : dulwich.objects.Tree
        Tree object to mount on.
    components : list of str
        A list of strings (or bytes) of subpaths (e.g. ['foo', 'bar'] is equivalent to
        '/foo/bar').
    obj : dulwich.objects.Blob or dulwich.objects.Tree or None
        Object to mount. If None, removes the object found at path and prunes the tree
        downwards.

    Returns
    -------
    list of dulwich.objects:
        A list of new entities that need to be added to the object store, where the last
        one is the new tree.

    """
    # pattern-matching:
    if len(components) == 1:
        if isinstance(obj, Blob):
            mode: Optional[int] = 0o100644
        elif isinstance(obj, Tree):
            mode = 0o040000
        elif obj is None:
            mode = None
        else:
            raise TypeError("Can only mount Blobs or Trees")
        name = components[0]

        if mode is not None:
            assert obj is not None
            tree[name] = mode, obj.id
            return [tree]
        if name in tree:
            del tree[name]
        return [tree]
    elif len(components) > 1:
        a, bc = components[0], components[1:]
        if a in tree:
            a_tree = repo[tree[a][1]]
            if not isinstance(a_tree, Tree):
                a_tree = Tree()
        else:
            a_tree = Tree()
        res = _on_tree(repo, a_tree, bc, obj)
        a_tree_new = res[-1]

        if a_tree_new.items():
            tree[a] = 0o040000, a_tree_new.id
            return res + [tree]

        # tree is empty
        if a in tree:
            del tree[a]
        return [tree]
    else:
        raise ValueError("Components can't be empty.")


class GitCommitStore(KeyValueStore):
    """Store using git.

    Parameters
    ----------
    repo_path : str
        Path to the repository.
    branch : bytes, optional, default = b"master"
        Branch to use.
    subdir : bytes, optional, default = b""
        Subdirectory of the repository to use.

    """

    AUTHOR = f"GitCommitStore (minimalkv {__version__}) <>"
    TIMEZONE = None

    def __init__(self, repo_path: str, branch: bytes = b"master", subdir: bytes = b""):
        self.repo = Repo(repo_path)
        self.branch = branch

        # cleans up subdir, to a form of 'a/b/c' (no duplicate, leading or
        # trailing slashes)
        self.subdir = re.sub("#/+#", "/", subdir.decode("ascii").strip("/"))

    @property
    def _subdir_components(self) -> list[bytes]:
        return [c.encode("ascii") for c in self.subdir.split("/")]

    def _key_components(self, key: str) -> list[bytes]:
        return [c.encode("ascii") for c in key.split("/")]

    @property
    def _refname(self):
        return b"refs/heads/" + self.branch

    def _create_top_commit(self):
        # get the top commit, create empty one if it does not exist
        commit = Commit()

        # commit metadata
        author = self.AUTHOR.encode("utf8")
        commit.author = commit.committer = author
        commit.commit_time = commit.author_time = int(time.time())

        if self.TIMEZONE is not None:
            tz = self.TIMEZONE
        else:
            tz = time.timezone if (time.localtime().tm_isdst) else time.altzone
        commit.commit_timezone = commit.author_timezone = tz
        commit.encoding = b"UTF-8"

        return commit

    def _delete(self, key: str) -> None:
        try:
            commit = self.repo[self._refname]
            tree = self.repo[commit.tree]
        except KeyError:
            return  # not-found key errors are ignored

        commit = self._create_top_commit()
        objects_to_add = []

        components = self._key_components(key)
        if self.subdir:
            components = self._subdir_components + components

        res = _on_tree(self.repo, tree, components, None)
        objects_to_add.extend(res)
        tree = res[-1]

        commit.tree = tree.id
        commit.message = ("Deleted key {}".format(self.subdir + "/" + key)).encode(
            "utf8"
        )

        objects_to_add.append(commit)

        for obj in objects_to_add:
            self.repo.object_store.add_object(obj)

        self.repo.refs[self._refname] = commit.id

    def _get(self, key: str) -> bytes:
        # might raise key errors, except block corrects param
        try:
            commit = self.repo[self._refname]
            tree = self.repo[commit.tree]
            fn = self.subdir + "/" + key
            _, blob_id = tree.lookup_path(self.repo.__getitem__, fn.encode("ascii"))
            blob = self.repo[blob_id]
        except KeyError as e:
            raise KeyError(key) from e

        return blob.data

    def iter_keys(self, prefix: str = "") -> Iterator[str]:  # noqa D
        try:
            commit = self.repo[self._refname]
            tree = self.repo[commit.tree]

            if self.subdir:
                tree = self.repo[
                    tree.lookup_path(
                        self.repo.__getitem__, self.subdir.encode("ascii")
                    )[1]
                ]
        except KeyError:
            pass
        else:
            for o in self.repo.object_store.iter_tree_contents(
                tree.sha().hexdigest().encode("ascii")
            ):
                if o.path.decode("ascii").startswith(prefix):
                    yield o.path.decode("ascii")

    def _open(self, key: str) -> BinaryIO:
        return BytesIO(self._get(key))

    def _put_file(self, key: str, file: BinaryIO) -> str:
        # FIXME: it may be worth to try to move large files directly into the
        #        store here
        return self._put(key, file.read())

    def _put(self, key: str, data: bytes) -> str:
        commit = self._create_top_commit()
        commit.message = ("Updated key {}".format(self.subdir + "/" + key)).encode(
            "utf8"
        )

        blob = Blob.from_string(data)

        try:
            parent_commit = self.repo[self._refname]
        except KeyError:
            # branch does not exist, start with an empty tree
            tree = Tree()
        else:
            commit.parents = [parent_commit.id]
            tree = self.repo[parent_commit.tree]

        objects_to_add = [blob]

        components = self._key_components(key)
        if self.subdir:
            components = self._subdir_components + components
        res = _on_tree(self.repo, tree, components, blob)
        objects_to_add.extend(res)

        commit.tree = res[-1].id
        objects_to_add.append(commit)

        # add objects
        for obj in objects_to_add:
            self.repo.object_store.add_object(obj)

        # update refs
        self.repo.refs[self._refname] = commit.id

        return key



# Copyright (c) QuantCo 2024-2024
# SPDX-License-Identifier: LicenseRef-QuantCo

from collections.abc import Coroutine
from datetime import datetime
from logging import getLogger
from typing import Any, Callable, Optional, TypedDict

from aiobotocore.credentials import (
    AioRefreshableCredentials,
    CredentialResolver,
    create_assume_role_refresher,
)
from aiobotocore.session import AioSession
from botocore.credentials import CredentialProvider
from dateutil import tz  # type: ignore[import-untyped]

logger = getLogger(__name__)


class BotoCredentials(TypedDict):
    access_key: str
    secret_key: str
    token: str
    expiry_time: str


class RefreshableAssumeRoleProvider(CredentialProvider):
    """A credential provider that loads self-refreshing credentials using 'sts.assume_role'.

    The credentials are obtained in the '_refresh' method using the given `sts_session`
    and 'assume_role_params' to assume a role.
    Whenever a refresh is needed, the credentials are obtained
    again by assuming the role again through the `sts_session`.

    This means that if the given `sts_session` is expired, the refreshable credentials
    also won't be able to refresh themselves.
    """

    METHOD = "refreshable-assume-role"

    def __init__(
        self,
        sts_session: AioSession,
        assume_role_params: dict,
        advisory_timeout: Optional[int] = None,
        mandatory_timeout: Optional[int] = None,
    ):
        self.sts_session = sts_session
        self._assume_role_params = assume_role_params
        self.advisory_timeout = advisory_timeout
        self.mandatory_timeout = mandatory_timeout

        super().__init__()

    async def load(self) -> AioRefreshableCredentials:
        refresh = self.create_refresh()

        refreshable_credentials = AioRefreshableCredentials.create_from_metadata(
            metadata=await refresh(),
            refresh_using=refresh,
            method=self.METHOD,
            advisory_timeout=self.advisory_timeout,
            mandatory_timeout=self.mandatory_timeout,
        )

        return refreshable_credentials

    def create_refresh(self) -> Callable[[], Coroutine[Any, Any, BotoCredentials]]:
        async def _refresh() -> BotoCredentials:
            logger.info("Refreshing credentials")
            sts_client = self.sts_session.create_client("sts")
            refresh = create_assume_role_refresher(sts_client, self._assume_role_params)

            credentials: BotoCredentials = await refresh()
            to_zone = tz.tzlocal()
            local_expiry_time = (
                datetime.strptime(credentials["expiry_time"], "%Y-%m-%dT%H:%M:%S%Z")
                .replace(tzinfo=tz.tzutc())
                .astimezone(to_zone)
            )
            logger.info(
                f"""Refreshed credentials with access key '{credentials["access_key"]}' and expiry time '{local_expiry_time}'."""
            )
            return credentials

        return _refresh


def create_aio_session_w_refreshable_credentials(
    access_key: str,
    secret_key: str,
    token: Optional[str],
    assume_role_params: dict,
    advisory_timeout: Optional[int] = None,
    mandatory_timeout: Optional[int] = None,
):
    """Create an aio session that knows how to refresh its credentials using 'sts.assume_role'.

    The 'aws_credentials' are used to assume a role (specified by assume_role_params)
    and if refresh is needed, the session will use the credentials to assume the role
    again.

    This means that if the given 'aws_credentials' are expired, the returned session
    also won't be able to refresh its credentials.
    """
    # This session is used to provide the refresh mechanism
    sts_session = AioSession()
    sts_session.set_credentials(
        access_key=access_key,
        secret_key=secret_key,
        token=token,
    )

    # Actual session with refreshable credentials
    session = AioSession()
    resolver = CredentialResolver(
        providers=[
            RefreshableAssumeRoleProvider(
                sts_session, assume_role_params, advisory_timeout, mandatory_timeout
            )
        ]
        # Only offer 1 way of retrieving credentials
        # => Do not need to worry about "side-effects" anymore
    )
    session.register_component(
        "credential_provider", resolver
    )  # "registering" replaces the current credential provider with our resolver

    return session
